// pfodVC820MeterParser.cpp
// Parser for RS232 reading from TEKPOWER TP4000ZC Digitech QM1538, Digitech QM1537, 
// Digitek DT-9062, Digitek INO2513, Digitech QM1462, 
// PeakTech 3330, Tenma 72-7745, Uni-Trend UT30A, Uni-Trend UT30E, Uni-Trend UT60E, 
// Voltcraft VC 820, Voltcraft VC 840

#include "pfodVC820MeterParser.h"
/**
pfodVC820MeterParser for Arduino

(c)2016 Forward Computing and Control Pty. Ltd.
www.forward.com.au
This code may be freely used for both private and commercial use.
Provide this copyright is maintained.
*/

//#define DEBUG
// need to call setDebugOut(&Serial); as well as un-commenting #define DEBUG above to get debug output

/**
    +----------------------------<br>
    Each measurement is sent as a packet of 14 bytes. This encodes the<br>
    LCD display state, not the numeric measurement data.<br>
    <br>
    Each byte has the high order nibble set to its sequence number,<br>
    starting from 1. A packet looks something like -<br>
    <br>
    15 27 3D 47 5D 67 7D 8F 9D A0 B0 C0 D4 E8<br>
    <br>
    The low order nibble contains the display data. As far as I am able
    to<br>
    determine, each bit corresponds to a display element, and is set to<br>
    '1' if the element is on, or '0' if it is off.<br>
    <br>
    The bits correspond to display elements as follows -<br>
    <br>
    Byte 1x - bit 0001 -> RS232 (?)<br>
    - bit 0010 -> Auto<br>
    - bit 0100 -> DC (not displayed)<br>
    - bit 1000 -> AC<br>
    <br>
    The next 8 bytes encode the seven segment display values for the<br>
    four digits. These are encoded in two consecutive nibbles.<br>
    <br>
    Byte 2x & 3y correspond to digit 1 (leftmost)<br>
    Byte 4x & 5y correspond to digit 2<br>
    Byte 6x & 7y correspond to digit 3<br>
    Byte 8x & 9y correspond to digit 4<br>
    <br>
    Taking the two nibbles together, the segment values are encoded -<br>
    <br>
    bit 0000 0001 - top right<br>
    bit 0000 0010 - centre<br>
    bit 0000 0100 - bottom right<br>
    bit 0000 1000 - bottom<br>
    bit 0001 0000 - top<br>
    bit 0010 0000 - top left<br>
    bit 0100 0000 - bottom left<br>
    bit 1000 0000 - minus sign for digit 1, leading decimal point for other
    digits<br>
    <br>
    Byte Ax - bit 0001 -> diode indicator<br>
    - bit 0010 -> kilo indicator<br>
    - bit 0100 -> nano (or Farrad?)<br>
    - bit 1000 -> micro indicator<br>
    <br>
    Byte Bx - bit 0001 -> speaker indicator<br>
    - bit 0010 -> mega indicator<br>
    - bit 0100 -> % indicator<br>
    - bit 1000 -> milli indicator<br>
    <br>
    Byte Cx - bit 0001 -> Hold indicator<br>
    - bit 0010 -> Rel indicator<br>
    - bit 0100 -> Ohm indicator<br>
    - bit 1000 -> Farrad (or nano?)<br>
    <br>
    Byte Dx - bit 0001 -> unknown or unused<br>
    - bit 0010 -> Hz indicator<br>
    - bit 0100 -> V indicator<br>
    - bit 1000 -> A indicator<br>
    <br>
    Byte Ex - bit 0001 -> unknown or unused<br>
    - bit 0010 -> unknown or unused<br>
    - bit 0100 -> degC indicator<br>
    - bit 1000 -> unknown or unused<br>
    for all other measurements.<br>
*/

pfodVC820MeterParser::pfodVC820MeterParser(void) {
  init();
  debugOut = NULL;
  clearOutByteArray();
}

void pfodVC820MeterParser::setDebugOut(Print *_debugOut) {
#ifdef DEBUG
  debugOut = _debugOut;
#endif
}

void pfodVC820MeterParser::init() {
  measurementType = NO_READING;
  measurementScaling = NO_SCALING;
  autoSet = false;
  acSet = false; // true for AC measurements
  relativeSet = false;
  overflowSet = false;
  negSet = false;
  holdSet = false;
  diodeSet = false;
  haveNonZeroDigit = false;
  reading[0] = '\0';
  reading[MAX_READING_LEN] = '\0'; // array is MAX_READING_LEN+1 long
  strReading[0] = '\0';
  strReading[MAX_STR_READING_LEN] = '\0'; // array is MAX_STR_READING_LEN+1 long
}

// clear out byte array
void pfodVC820MeterParser::clearOutByteArray() {
  for (int i = 0; i < MAX_BYTES; i++) {
    byteArray[i] = 0;
  }
}
bool pfodVC820MeterParser::haveReading() {
  bool completeReading = false;
  int byteCounter = 0;
  while (meterInput->available() && (!completeReading) && (byteCounter<14)) {  // process any buffered bytes
  	byteCounter++; // make sure we don't loop here forever regardless
    int b = meterInput->read();
    if (b < 0) {
    	return false; // end of data
    }
    // sync on byte 1
    int byteIdx = (0x0f & (b >> 4)) - 1;
    if (byteIdx == 0) {  // start new reading
      clearOutByteArray();
    }
    if ((byteIdx >= 0) && (byteIdx < MAX_BYTES)) {
      byteArray[byteIdx] = b;
      if (byteIdx == (MAX_BYTES - 1)) {
      // found last byte
#ifdef DEBUG
      if (debugOut) {
          for (int i = 0; i < MAX_BYTES; i++) {
            debugOut->print(byteArray[i], HEX);  Serial.print(' ');
          }
          debugOut->println();
        }
#endif
        if (checkFormat()) {
          completeReading = true;
#ifdef DEBUG
          if (debugOut) {
            debugOut->println(F("Valid sequence"));
          }
#endif
        } else {
#ifdef DEBUG
          if (debugOut) {
            debugOut->println(F("Invalid sequence"));
          }
#endif
        }
      }
    }
  }
  if (!completeReading) {
    return false;
  }
  // else parse complete reading
  init(); // sets invalid
  parseByte0ACDC(byteArray); // <<< call this first so that when parsing HzVA the correct V and A setting is set

  size_t readingLeft = MAX_READING_LEN;
  readingLeft = parseByteDigit(byteArray, reading, readingLeft, 0);
  readingLeft = parseByteDigit(byteArray, reading, readingLeft, 1);
  readingLeft = parseByteDigit(byteArray, reading, readingLeft, 2);
  readingLeft = parseByteDigit(byteArray, reading, readingLeft, 3);
  if (isOverflow()) { // set return strings
    reading[0] = '\0';
    readingLeft = MAX_READING_LEN;
    safe_ncat(reading, getFstr(F("OL "), fStr), readingLeft);
    readingLeft =  MAX_STR_READING_LEN;
    safe_ncat(strReading, getFstr(F("OL "), fStr), readingLeft);
  }
  parseByte0AutoRS232(byteArray);
  parseRelOhmFarrad(byteArray);
  parseHzVA(byteArray);
  parseDegC(byteArray);
  parseMegaMilliPecent(byteArray);
  parseKiloNanoMicro(byteArray);

  if (!isValid()) {
    floatReading = 0.0;
    measurementScaling = NO_SCALING; 
  } else {
    if (overflowSet) {
      floatReading = atof(getFstr(F("inf"), fStr));
    } else {
      readingLeft =  MAX_STR_READING_LEN;
      readingLeft =  safe_ncat(strReading, reading, readingLeft);
      readingLeft =  safe_ncat(strReading, getScalingAsNumberStr(), readingLeft);
      floatReading = atof(reading);
    }
  }
#ifdef DEBUG
  if (debugOut) {
    debugOut->println();
    debugOut->print(getDigits()); debugOut->print(getScalingAsStr()); debugOut->print(getTypeAsStr()); debugOut->print(F(" = ")); debugOut->print(getAsFloat());
    debugOut->println();
    debugOut->print(getAsStr());  debugOut->print(getTypeAsStr()); debugOut->print(F(" = ")); debugOut->print(getAsFloat());
    debugOut->println();
  }
#endif

  if (isValid()) {
    return true;
  }
  //else invalid for some reason
  return false;
}

// check for top nibble == 1,2...,14
bool pfodVC820MeterParser::checkFormat() {
  for (int i = 0; i < MAX_BYTES; i++) {
    if (getByteIdx(byteArray[i]) != i) {
      return false;
    }
  }
  return true; // valid
}


void pfodVC820MeterParser::connect(Stream * input) {
  init();
  clearOutByteArray();
  meterInput = input;
}

bool pfodVC820MeterParser::isRelative() {
  return relativeSet;
}
bool pfodVC820MeterParser::isAuto() {
  return autoSet;
}
bool pfodVC820MeterParser::isHold() {
  return holdSet;
}

bool pfodVC820MeterParser::isOverflow() {
  return overflowSet;
}

const char* pfodVC820MeterParser::getDigits() {
  if (isValid()) {
    return reading;
  } else {
    return emptyStr; // invalid
  }
}

const char *pfodVC820MeterParser::getAsStr() { // reading as string with units i.e. KOhms, nF, mV (no AC or DC text)
  if (!isValid()) {
    return emptyStr;
  }
  return strReading;
}

int pfodVC820MeterParser::getType() {
  return measurementType;
}

// rtn must be large enought to hold largest str
const char* pfodVC820MeterParser::getFstr(const __FlashStringHelper *ifsh, char*rtn) {
  PGM_P p = reinterpret_cast<PGM_P>(ifsh);
  int idx = 0;
  while (1) {
    unsigned char c = pgm_read_byte(p++);
    rtn[idx++] = c;
    if (c == 0) break;
  }
  return (const char*)rtn;
}

const char* pfodVC820MeterParser::getTypeAsStr() {
  switch (measurementType) {
    case VOLTS_AC:
      return getFstr(F("V AC"), fStr);
    case VOLTS_DC:
      return getFstr(F("V DC"), fStr);
    case AMPS_AC:
      return getFstr(F("A AC"), fStr);
    case AMPS_DC:
      return getFstr(F("A DC"), fStr);
    case OHMS:
      return getFstr(F("ohms"), fStr);
    case HZ:
      return getFstr(F("Hz"), fStr);
    case DUTY_CYCLE_PERCENT:
      return getFstr(F("%"), fStr);
    case FARADS:
      return getFstr(F("F"), fStr);
    case DEGS_C:
      return getFstr(F("C"), fStr);
    case NO_READING: // same as default
    default:
      return emptyStr;
  }
}

const char* pfodVC820MeterParser::getTypeAsUnicode() {
  switch (measurementType) {
    case OHMS:
      return getFstr(F("\316\251"), fStr);
    case DEGS_C:
      return getFstr(F("\342\204\203"), fStr);
    default:
      return getTypeAsStr();
  }
}

float pfodVC820MeterParser::getScaling() {
  if (!isValid()) {
    return 1;
  }
  switch (measurementScaling) {
    case KILO:
      return 1e3;
    case MEGA:
      return 1e6;
    case MILLI:
      return 1e-3;
    case MICRO:
      return 1e-6;
    case NANO:
      return 1e-9;
    case NO_SCALING:
    default:
      return 1.0;
  }
}

const char* pfodVC820MeterParser::getScalingAsNumberStr() {
  if (!isValid()) {
    return emptyStr;
  }
  switch (measurementScaling) {
    case KILO:
      return getFstr(F("e3"), fStr);
    case MEGA:
      return getFstr(F("e6"), fStr);
    case MILLI:
      return getFstr(F("e-3"), fStr);
    case MICRO:
      return getFstr(F("e-6"), fStr);
    case NANO:
      return getFstr(F("e-9"), fStr);
    case NO_SCALING:
    default:
      return emptyStr;
  }
}

const char* pfodVC820MeterParser::getScalingAsStr() {
  if (!isValid()) {
    return emptyStr;
  }
  switch (measurementScaling) {
    case KILO:
      return getFstr(F("K"), fStr);
    case MEGA:
      return getFstr(F("M"), fStr);
    case MILLI:
      return getFstr(F("m"), fStr);
    case MICRO:
      return getFstr(F("u"), fStr);
    case NANO:
      return getFstr(F("n"), fStr);
    case NO_SCALING:
    default:
      return emptyStr;
  }
}

bool pfodVC820MeterParser::isValid() {
  return (measurementType != NO_READING);
}
float pfodVC820MeterParser::getAsFloat() { // reading as float
  if (isValid()) {
    return floatReading * getScaling();
  } else {
    return 0.0; // return 0 if invalid
  }
}


size_t pfodVC820MeterParser::safe_addChar(char* dest, const char c, size_t spaceRemaining) {
  if (spaceRemaining < 1) {
    return spaceRemaining; // do nothing
  }
  size_t lenIn = strlen(dest);
  dest[lenIn] = c;
  dest[lenIn + 1] = '\0';
  return (spaceRemaining - 1);
}

size_t pfodVC820MeterParser::safe_ncat(char* dest, const char* src, size_t spaceRemaining) {
  if (spaceRemaining == 0) {
    return spaceRemaining; // do nothing
  }
  size_t lenIn = strlen(dest);
  dest = strncat(dest, src, spaceRemaining - 1); // allow for terminating null
  return (spaceRemaining - (strlen(dest) - lenIn));
}

// first byte
//    Byte 1x - bit 0001 -> RS232 (?)<br>
//    - bit 0010 -> Auto<br>
//    - bit 0100 -> DC (not displayed)<br>
//    - bit 1000 -> AC<br>
/**
   size_t spaceRemaining is the space remaining including the terminating null!!
   so define rtn as char rtn[MAX_SIZE+1]; // to allow for terminating null
*/
void pfodVC820MeterParser::parseByte0AutoRS232(byte * barray) {
  byte b = barray[0];
  if ((b & 0x01) != 0) {
    // skip this as we know it is RS232
    //spaceRemaining = safe_ncat(rtn, " RS232 ", spaceRemaining); // allow for terminating null
  }
  if ((b & 0x02) != 0) {
    autoSet = true;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" Auto "));
    }
#endif
  }
}


// first byte
//    Byte 1x - bit 0001 -> RS232 (?)<br>
//    - bit 0010 -> Auto<br>
//    - bit 0100 -> DC (not displayed)<br>
//    - bit 1000 -> AC<br>
void pfodVC820MeterParser::parseByte0ACDC(byte * barray) {
  byte b = barray[0];
  if ((b & 0x04) != 0) {
    acSet = false;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" DC "));
    }
#endif
  }
  if ((b & 0x08) != 0) {
    acSet = true;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" AC "));
    }
#endif
  }
}


// byte idx = 12
//    Byte Dx - bit 0001 -> unknown or unused<br>
//    - bit 0010 -> Hz indicator<br>
//    - bit 0100 -> V indicator<br>
//    - bit 1000 -> A indicator<br>
void pfodVC820MeterParser::parseHzVA(byte * barray) {
  byte b = barray[12];
  if ((b & 0x01) != 0) {
    //indicator skip
  }
  if ((b & 0x02) != 0) {
    measurementType = HZ;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" Hz "));
    }
#endif
  }
  if ((b & 0x04) != 0) {
    if (acSet) {
      measurementType = VOLTS_AC;
    } else {
      measurementType = VOLTS_DC;
    }
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" V "));
    }
#endif
  }
  if ((b & 0x08) != 0) {
    if (acSet) {
      measurementType = AMPS_AC;
    } else {
      measurementType = AMPS_DC;
    }
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" A "));
    }
#endif
  }
}

//    Byte Ex
//    - bit 0001 -> unknown or unused<br>
//    - bit 0010 -> unknown or unused<br>
//    - bit 0100 -> degC indicator<br>
//    - bit 1000 -> unknown or unused<br><br>
void pfodVC820MeterParser::parseDegC(byte * barray) {
  byte b = barray[13];
  //  if ((b & 0x01) != 0) {
  //    //indicator skip
  //  }
  //  if ((b & 0x02) != 0) {
  //    //indicator skip
  //  }
  if ((b & 0x04) != 0) {
    measurementType = DEGS_C;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" degC "));
    }
#endif
  }
  //  if ((b & 0x08) != 0) {
  //    //indicator skip
  //  }
}

//    Byte Cx - bit 0001 -> Hold indicator<br>
//    - bit 0010 -> Rel indicator<br>
//    - bit 0100 -> Ohm indicator<br>
//    - bit 1000 -> Farrad (or nano?)<br>
void pfodVC820MeterParser::parseRelOhmFarrad(byte * barray) {
  byte b = barray[11];
  if ((b & 0x01) != 0) {
    holdSet = true;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" Hold "));
    }
#endif
  }
  if ((b & 0x02) != 0) {
    relativeSet = true;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" Rel "));
    }
#endif
  }
  if ((b & 0x04) != 0) {
    measurementType = OHMS;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" Ohms "));
    }
#endif
  }
  if ((b & 0x08) != 0) {
    measurementType = FARADS;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F(" F "));
    }
#endif
  }
}

//    Byte Bx - bit 0001 -> speaker indicator<br>
//    - bit 0010 -> mega indicator<br>
//    - bit 0100 -> % indicator<br>
//    - bit 1000 -> milli indicator<br>
void pfodVC820MeterParser::parseMegaMilliPecent(byte * barray) {
  byte b = barray[10];
  if ((b & 0x01) != 0) {
    //speaker indicator skip
  }
  if ((b & 0x02) != 0) {
    measurementScaling = MEGA;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F("e6 "));
    }
#endif
  }
  if ((b & 0x04) != 0) {
    measurementScaling = NO_SCALING;
    measurementType = DUTY_CYCLE_PERCENT;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F("% "));
    }
#endif
  }
  if ((b & 0x08) != 0) {
    measurementScaling = MILLI;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F("e-3 "));
    }
#endif
  }
}

//    Byte Ax - bit 0001 -> diode indicator<br>
//    - bit 0010 -> kilo indicator<br>
//    - bit 0100 -> nano (or Farrad?)<br>
//    - bit 1000 -> micro indicator<br>
void pfodVC820MeterParser::parseKiloNanoMicro(byte * barray) {
  byte b = barray[9];
  if ((b & 0x01) != 0) {
    diodeSet = true;
  }
  if ((b & 0x02) != 0) {
    measurementScaling = KILO;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F("e3 "));
    }
#endif
  }
  if ((b & 0x04) != 0) {
    measurementScaling = NANO;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F("e-9 "));
    }
#endif
  }
  if ((b & 0x08) != 0) {
    measurementScaling = MICRO;
#ifdef DEBUG
    if (debugOut) {
      debugOut->print(F("e-6 "));
    }
#endif
  }
}

//    <br>
//    The next 8 bytes encode the seven segment display values for the<br>
//    four digits. These are encoded in two consecutive nibbles.<br>
//    <br>
//    Byte 2x & 3y correspond to digit 1 (leftmost)<br>
//    Byte 4x & 5y correspond to digit 2<br>
//    Byte 6x & 7y correspond to digit 3<br>
//    Byte 8x & 9y correspond to digit 4<br>
//    <br>
//    Taking the two nibbles together, the segment values are encoded -<br>
//    <br>
//    bit 0000 0001 - top right<br>
//    bit 0000 0010 - centre<br>
//    bit 0000 0100 - bottom right<br>
//    bit 0000 1000 - bottom<br>
//    bit 0001 0000 - top<br>
//    bit 0010 0000 - top left<br>
//    bit 0100 0000 - bottom left<br>
//    bit 1000 0000 - minus sign for digit 1, leading decimal point for other digits<br>
//    <br>
/**
   Parse the 7 seg digit

   @param b1 first nibble
   @param b2 second nibble digits are 0 == 125, 1 = 5, 2 = 91, 3 = 31, 4 =
   39, 5 = 62, 6 = 126, 7 = 21, 8 = 127, 9 = 63 and L for 0.L L =
   @param rtn string rtn
   @param neg true if leading digit, i.e. 128 == neg else false if 128 ==
   decimal point

   digit is the 0 to 3 for first to 4th digit, if digit = 0 handle -ve sign
   @return
*/
size_t pfodVC820MeterParser::parseByteDigit(byte * barray, char* rtn, size_t spaceRemaining, int digitPos) {
  if (overflowSet) {
    return spaceRemaining;
  }
  bool setNeg = (digitPos == 0); // first digit has -ve sign rest may have dot.
  int b1_idx = 1 + (digitPos * 2);
  int b2_idx = 1 + b1_idx;
  byte b1 = barray[b1_idx] & 0x0F;
  byte b2 = barray[b2_idx] & 0x0F;
  byte digit = (byte) (((byte) (b1 << 4)) + ((byte) b2));

  if ((byte) (b1 & 0x08) != 0) {
    // neg or point
    if (setNeg) {
      negSet = true;
      spaceRemaining = safe_addChar(rtn, '-', spaceRemaining);
    } else {
      if (!haveNonZeroDigit) {
        // add a 0 before this .
        haveNonZeroDigit = true;
        spaceRemaining = safe_addChar(rtn, '0', spaceRemaining);
      }
      spaceRemaining = safe_addChar(rtn, '.', spaceRemaining);
    }
  }
  b1 = (byte) (b1 & 0x07); // remove neg or decimal point
  digit = (byte) (((byte) (b1 << 4)) + ((byte) b2));
  if (digit == 0) {
    return spaceRemaining; // no digit
  }
  bool haveDigit = true;
  // check for overload
  if (digit == (byte) 104) {
    overflowSet = true;
    return spaceRemaining; // stop here
  }
  if (digit == (byte) 125) {
    if (haveNonZeroDigit) {
      // add this zero
      spaceRemaining = safe_addChar(rtn, '0', spaceRemaining);
    }
    haveDigit = false; // all others true
  } else if (digit == (byte) 5) {
    spaceRemaining = safe_addChar(rtn, '1', spaceRemaining);
  } else if (digit == (byte) 91) {
    spaceRemaining = safe_addChar(rtn, '2', spaceRemaining);
  } else if (digit == (byte) 31) {
    spaceRemaining = safe_addChar(rtn, '3', spaceRemaining);
  } else if (digit == (byte) 39) {
    spaceRemaining = safe_addChar(rtn, '4', spaceRemaining);
  } else if (digit == (byte) 62) {
    spaceRemaining = safe_addChar(rtn, '5', spaceRemaining);
  } else if (digit == (byte) 126) {
    spaceRemaining = safe_addChar(rtn, '6', spaceRemaining);
  } else if (digit == (byte) 21) {
    spaceRemaining = safe_addChar(rtn, '7', spaceRemaining);
  } else if (digit == (byte) 127) {
    spaceRemaining = safe_addChar(rtn, '8', spaceRemaining);
  } else if (digit == (byte) 63) {
    spaceRemaining = safe_addChar(rtn, '9', spaceRemaining);
  } else {
    measurementType = NO_READING;
  }
  if (!haveNonZeroDigit) {
    // update from last digit
    haveNonZeroDigit = haveDigit;
  }
  return spaceRemaining;
}


